//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup Tool "FUNSSEL" developed for the 
// Simulator League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public 
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "Base.h"
#include <Config.h>
#include <constants.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <errno.h>

// from soccerserver
namespace Soccer {
#include <param.h>
#include <types.h>
}

#define FORCE_CORE

//#define DEBUG

// be FUNSSEL AND console driven FUNSSELMonitor
//#define CONSOLE

#include "FunsselStrings.h"

#ifdef CONSOLE
#include "Socket.h"
#endif
// cwz-build-command: make FUNSSEL

int Base::strncmp(const string &s, const char *str, int len)
{
  return ::strncmp(s.data(), str, len);
}

//////////////////////////////////////////////////////////////

Base::Base(const string &server_host, const string &_logfile)
  throw (Exception) :
  fsocket(Config::getValueAsInt("proxy_socket"),
	  server_host.c_str(),
	  Config::getValueAsInt("server_socket"),
	  MAX_SERVER_MSG_SIZE),
  fsocket_ol(Config::getValueAsInt("proxy_olcoach_socket"),
	     server_host.c_str(),
	     Config::getValueAsInt("server_olcoach_socket"),
	     MAX_SERVER_MSG_SIZE),
  fsocket_cc(Config::getValueAsInt("proxy_coach_socket"),
	     server_host.c_str(),
	     Config::getValueAsInt("server_coach_socket"),
	     MAX_SERVER_MSG_SIZE),
  mode(Playback),
  srv_time(-1),
  max_run(6000),
  max_delay(Config::getValueAsInt("max_delay")),
  poll_pending_time(-2),
  messages(_logfile),
  finished(false),
  log_board(!Config::getValueAsInt("filter_board_log")),
  msg_board(!Config::getValueAsInt("filter_board_audio")),
  log_commands(!Config::getValueAsInt("filter_commands"))
{
  clients.erase(clients.begin(), clients.end());
  for (int i = 0; i < NumberSActors; i++)
    clientIDOf[i] = -1;
}

Base::~Base()
{
  if (mode != Playback)
    server.kill();
  dropNonFunsselMon();
}

bool Base::sendCurMsg()
{
  SoccerMessage msg = messages.getNextMessage();
  while (msg.isValid() && (msg.recver != FunsselMonitor)) {
    msg = messages.getNextMessage();
  }
  if (!msg.isValid()) {
    poll_pending_time = messages.getCurrentFetchTime();
    sendFMsg(FunsselMonitor, "dispeol");
    return true;
  }

  poll_pending_time = -2;
  sendFMon(msg);
  return true;
}

///////////////////////////////////////////////////// sockets

Base::SoccerClient *Base::getMessage(FilterMessage &msg) throw (Exception)
{
  SocketDesc desc = S_None;
  msg = fsocket.getNextMessage();
  if (msg.isValid())
    desc = S_Primary;
  else {
    msg = fsocket_ol.getNextMessage();
    if (msg.isValid())
      desc = S_OnlineCoach;
    else {
      msg = fsocket_cc.getNextMessage();
      if (msg.isValid())
	desc = S_OfflineCoach;
      else
	// no message
	return NULL;
    }
  }

  SoccerClient &clnt = findClient(msg.client);
  clnt.port = desc;
  return &clnt;
}

void Base::sendMessage(FilterMessage &msg) throw (Exception)
{
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (it->clientID == msg.client)
      break;
  if (it == clients.end())
    throw Exception("base: send to unknown client");
  switch (it->port) {
  case S_OnlineCoach:
    fsocket_ol.send(msg);
    break;
  case S_OfflineCoach:
    fsocket_cc.send(msg);
    break;
  default:
    fsocket.send(msg);
  }
}

void Base::dropNonFunsselMon() throw (Exception)
{
  // first temporary throw out all clients that are not
  // FunsselMonitors. They would not survive a server
  // restart anyway
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++) {
    if (it->actor != FunsselMonitor) {
      switch (it->port) {
      case S_Primary:
	fsocket.drop(it->clientID);
	break;
      case S_OnlineCoach:
	fsocket_ol.drop(it->clientID);
	break;
      case S_OfflineCoach:
	fsocket_cc.drop(it->clientID);
	break;
      default:;
      }
      // and declare unknown
      it->actor = NoOne;
    }
  }

  for (int i = 0; i < NumberSActors; i++) {
    // funssel monitor must not be thrown out
    if (i != FunsselMonitor)
      clientIDOf[i] = -1;
  }
}

///////////////////////////////////////// act

bool Base::act()
{
  FilterMessage msg("", -1, FilterMessage::FromServer);
  SoccerClient *clntP = getMessage(msg);
  if (clntP == NULL) {
    return false;
  }
  SoccerClient &clnt(*clntP);
  if (clnt.actor == NoOne) {
    // try to determine client type
    switch (clnt.port) {
    case S_OfflineCoach:
      // offline coach is easy to determine: by socket
      identify(clnt, OfflineCoach);
      break;
    case S_OnlineCoach:
      if (msg.fromServer()) {
	SoccerActor actor = coachSide(msg.message);
	if (actor != NoOne)
	  identify(clnt, actor);
	else
	  cerr << "FUNSSEL warning: cannot determine coach side (a message was lost)"
	       << endl;
      }
      break;
    case S_Primary:
      if (msg.fromServer()) {
	//player numbers are declared by server
	SoccerActor actor = playerSide(msg.message);
	if (actor != NoOne)
	  identify(clnt, actor);	  
      }
      else {
	// monitors declare themselves
	if (!strncmp(msg.message, "(disp", 5))
	  identify(clnt, SoccerMonitor);
	else if (!strncmp(msg.message, FILTER_STRING " (dispinit))",
			  FILTER_STRING_LEN + 12))
	  identify(clnt, FunsselMonitor);
      }
      break;
    default:
      throw Exception("socket id illegal");
    }
  }

  // convert into SoccerMessage
  SoccerMessage s_msg(0, "", NoOne, NoOne);

  if (msg.fromServer()) {
    // messages from the server are no problem
    if (mode == Playback) {
      cerr << "server messages in logplay mode?" << endl;
      return true;
    }
    adjustTime(msg.message);

    // see wether the message shall be stored
    bool store = log_commands;
    
    Soccer::dispinfo_t2 *d = (Soccer::dispinfo_t2 *)(msg.message.data());
    switch (ntohs(d->mode)) {
    case MSG_MODE:
      switch (ntohs(d->body.msg.board)) {
      case MSG_BOARD:
	store = msg_board;
	break;
      case LOG_BOARD:
	store = log_board;
	break;
      default: ;
      }
      break;
    case SHOW_MODE:
    case PT_MODE:
    case PPARAM_MODE:
      store = true;
      break;
    case DRAW_MODE:
    case BLANK_MODE:
    case PM_MODE:
    case TEAM_MODE:
    case PARAM_MODE:
    default:
      store = false;
      break;
    }

    if (store) {
      s_msg.sender = SoccerServer;
      s_msg.recver = clnt.actor;
      s_msg.time   = srv_time;
      s_msg.durtn  = 1;
      s_msg.message = msg.message;
    }

    // forward original message, even if it is not stored
    if (clnt.actor != FunsselMonitor)
      sendMessage(msg);
  }
  else {
    // now it gets tricky, thanks to the proxy protocol
    if (!strncmp(msg.message, FILTER_STRING, FILTER_STRING_LEN)) {
      // message has proxy form

      // filter message are only forwarded from known clients
      if (clnt.actor == NoOne)
	return true;

      const char *c_msg = msg.message.c_str() + FILTER_STRING_LEN + 1;
      int length = msg.message.length() - FILTER_STRING_LEN - 1;
      if (length <= 1) {
	sendFMsg(clnt.actor, "error illegal proxy command");
	return true;
      }
      if (*c_msg == '(') {
	// a funssel special command
	if (!strncmp(c_msg,"(dispmsg", 8)) {
	  c_msg  += 8;
	  length -= 8;
	  // a protected message - unpack it
	  int len, n, i_recver, i_sender;
	  if (sscanf(c_msg + 8, "%d %d %d %d %d %n", &s_msg.time,
		     &s_msg.durtn, &i_sender, &i_recver,
		     &len, &n) != 6) {
	    sendFMsg(clnt.actor, "error illegal dispmsg");
	    return true;
	  }
	  c_msg  += n;
	  length -= n;
	  if (length < len) {
	    sendFMsg(clnt.actor, "error message too short");
	    return true;
	  }
	  if (s_msg.time == -2)
	    s_msg.time = srv_time;
	  s_msg.sender = (SoccerActor) i_sender;
	  s_msg.recver = (SoccerActor) i_recver;
	  s_msg.message.assign(c_msg, len);
	  // continue with the unpacked message
	  // forward if not for FunsselMonitor (polling)
	  // and is not for an ancient time step
	  if ((s_msg.recver != FunsselMonitor) && (s_msg.time == srv_time)) {
	    FilterMessage f_msg("", -1, FilterMessage::FromServer);
	    if (s_msg.recver == SoccerServer) {
	      f_msg.client  =  clientIDOf[s_msg.sender];
	      f_msg.fromSvr = true;
	    }
	    else {
	      f_msg.client  = clientIDOf[s_msg.recver];
	      f_msg.fromSvr = false;
	    }
	    f_msg.message = s_msg.message;
	    sendMessage(msg);
	  }
	}
	else {
	  parseSpecialCommand(clnt.actor, c_msg);
	  return true;
	}
      }
      else {
	// a normal proxy command - redirect to FunsselMonitor
	int len;
	s_msg.sender = clnt.actor;
	s_msg.recver = FunsselMonitor;
	s_msg.time   = srv_time;
	if (sscanf(c_msg, "%d %n", &s_msg.durtn, &len) != 1) {
	  sendFMsg(clnt.actor, "error illegal command");
	  return true;
	}
	s_msg.message = msg.message;
	// these messages are not directly forwarded
	// because the funssel uses the poll protocol
      }
    }
    else {
      // a normal message for the server
      if (log_commands) {
	s_msg.sender = clnt.actor;
	s_msg.recver = SoccerServer;
	s_msg.time   = srv_time;
	s_msg.durtn  = 1;
	s_msg.message = msg.message;
      }
      // and forward (original message)
      sendMessage(msg);
    }
  }

  // ignore messages for unknown clients (or invalided msgs)
  if ((mode != Playback) && (s_msg.sender != NoOne) &&
      (s_msg.recver != NoOne)) {
    if (poll_pending_time == messages.getCurrentFetchTime())
      // got a pending poll for the time we inserted this message
      sendCurMsg();
    else if (poll_pending_time != -2) {
      // poll for other time pending. Send dispeol
      poll_pending_time = -2;
      sendFMsg(FunsselMonitor, "dispeol");
    }
    if (!messages.insertMsg(s_msg)) {
      sendFMsg(FunsselMonitor, "(insert failed)");
    }
  }
#ifdef DEBUG
  else
    cerr << "ignoring message " << s_msg.sender << " " << s_msg.recver << endl;
#endif
  return true;
}

void Base::parseSpecialCommand(SoccerActor actor, const char *com)
{
  if (actor != FunsselMonitor) {
    char buf[32];
    // currently this proxy does not know anything about
    // proxy commands from players. Forward them to monitor
    sprintf(buf, "forward %d ", int(actor));
    sendFMsg(FunsselMonitor, string(buf) + com);
    return;
  }
  if (!strncmp(com, "(dispinit)", 10)) {
    sendFMsg(FunsselMonitor, "welcome");
    return;
  }
  if (!strncmp(com, "(quit)", 6)) {
    finished = true;
    return;
  }
  if (!strncmp(com, "(poll)", 6)) {
    sendCurMsg();
    return;
  }
  if (!strncmp(com, "(server start)", 14)) {
    startServer();
    int end = time(NULL) + 2;
    while (time(NULL) <= end)
      usleep(100000);
    sendFMsg(FunsselMonitor, "server started");
    return;
  }
  if (!strncmp(com, "(max_run", 8)) {
    if (strlen(com) <= 9) {
      sendFMsg(FunsselMonitor, "error illegal max_run");
      return;
    }
    max_run = atol(com + 9);
    sendFMsg(FunsselMonitor, "max_run ok");
    return;
  }
  if (!strncmp(com, "(wind ", 6)) {
    if (strlen(com) <= 8) {
      sendFMsg(FunsselMonitor, "error illegal fetch");
      return;
    }
    int sign = 0;
    if (com[6] == 'b')
      sign = -1;
    if (com[6] == 'f')
      sign =  1;
    int time = atol(com + ((sign != 0) ? 7 : 6));
    bool stat;
    if (sign != 0)
      stat = wind_relative((sign > 0) ? time : -time);
    else
      stat = wind(time);
    char buf[32];
    sprintf(buf, "wind %s %d", stat ? "ok" : "failed",
	    messages.getCurrentFetchTime());
    sendFMsg(FunsselMonitor, buf);
    return;
  }
  sendFMsg(actor, "error illegal command");
  return;
}

void Base::sendFMsg(SoccerActor actor, const string &msg) throw (Exception)
{
  if (clientIDOf[actor] < 0)
    return;

  char header[64];
  sprintf(header, FILTER_STRING " %d (", messages.getMaximalSeenTime());
  string tmp_msg = header + msg + "))";
  FilterMessage f_msg(tmp_msg, clientIDOf[actor], FilterMessage::FromServer);
  sendMessage(f_msg);
}

void Base::sendFMon(const SoccerMessage &s_msg) throw (Exception)
{
  char buf[256];
  sprintf(buf, "dispmsg %d %d %d %d %d ",
	  s_msg.time,   s_msg.durtn,
	  s_msg.sender, s_msg.recver,
	  s_msg.getLength());
  string msg(buf);
  msg += s_msg.message;
  sendFMsg(FunsselMonitor, msg);
}

void Base::startServer() throw (Exception)
{
  dropNonFunsselMon();

  mode = Running;

  server.restart(Config::getValue("server_command"),
		 Config::getValue("server_dir"));
  messages.openWrite();
}

void Base::stopServer() throw (Exception)
{
  server.stop();
  mode = Stopped;
}

void Base::contServer() throw (Exception)
{
  server.cont();
  mode = Running;
}

Base::SoccerClient &Base::findClient(int id)
{
  vector<SoccerClient>::iterator it;
  for (it = clients.begin(); it != clients.end(); it++)
    if (it->clientID == id)
      break;

  if (it == clients.end()) {
    // client's first connect
#ifdef DEBUG
    cerr << "got new client " << id << endl;
#endif
    SoccerClient tmp_clnt;
    tmp_clnt.clientID = id;
    tmp_clnt.actor    = NoOne;
    tmp_clnt.port     = S_None;
    clients.push_back(tmp_clnt);
    it = clients.end() - 1;
  } 
  return *it;
}

void Base::identify(SoccerClient &clnt, SoccerActor actor)
{
  clnt.actor = actor;
#ifdef DEBUG
  cout << "identify " << clnt.clientID << " as " << actor << endl;
#endif
  // first one is the only one FUNSSEL accepts
  if (clientIDOf[actor] == -1)
    clientIDOf[actor] = clnt.clientID;
}

bool Base::wind_relative(int t) throw (Exception)
{
  return wind(t + messages.getCurrentFetchTime());
}

bool Base::wind(int t) throw (Exception)
{
  bool res = messages.fetchTime(t);
  if ((mode == Stopped) &&
      (srv_time - messages.getCurrentFetchTime() <= max_delay) &&
      (srv_time < max_run))
    contServer();
  if (poll_pending_time != t) {
    // again, terminate pending poll
    poll_pending_time = -2;
    sendFMsg(FunsselMonitor, "dispeol");
  }
  return res;
}

void Base::adjustTime(const string &msg)
{
  int tmp_time = -2;
  // test binary message
  const Soccer::dispinfo_t2 *d =
    (const Soccer::dispinfo_t2 *)msg.data();
  if (d->mode == htons(SHOW_MODE))
    tmp_time = int((short)ntohs(d->body.show.time));
  else {
    // check if text message
    int timeInd = -1;
    if (!strncmp(msg, "(see ", 5))
      timeInd = 5;
    else if (!strncmp(msg, "(sense_body ", 12))
      timeInd = 12;
    else if (!strncmp(msg, "(hear ", 6))
      timeInd = 6;
    else if (!strncmp(msg, "(ok look ", 9))
      timeInd = 9;
    else
      // no time in here (e.g. error)
      return;
    tmp_time = atol(msg.substr(timeInd).c_str());
  }
  // ok, we found a time stamp
  // too old, maybe packet reordering. Ok 
  if (tmp_time <= srv_time)
    return;
  // again NODA's wonderful world. There is no time 0...
  if ((tmp_time != 1) && (tmp_time != srv_time + 1)) {
    cerr << "FUNSSEL warning: maybe lost packets? lt: "
	 << tmp_time << " " << srv_time << endl;
  }
  if (tmp_time != srv_time) {
    sendFMsg(FunsselMonitor, "server time changed");
  }

#ifdef DEBUG
  cout << "detected new time step" << endl;
#endif
  srv_time = tmp_time;
  if ((srv_time - messages.getCurrentFetchTime() > max_delay) ||
      (srv_time >= max_run))
    stopServer();
}

SoccerActor Base::coachSide(const string &msg)
{
  if (strncmp(msg, "(init ", 6))
    return NoOne;

#ifdef DEBUG
  cerr << "coach got init msg" << endl;
#endif
  
  bool left = (msg[6] == 'l');
  
#ifdef DEBUG
  if (!left && (msg[6] != 'r'))
    cout << "left/right key failed:" << msg << " <" << msg[6] << ">" << endl;
#endif

#ifdef DEBUG
  cerr << "got olcoach " << left << endl;
#endif
  
  return (left) ? CoachL : CoachR;
}

SoccerActor Base::playerSide(const string &msg)
{
  if (strncmp(msg, "(init ", 6))
    return NoOne;

#ifdef DEBUG
  cerr << "player got init msg" << endl;
#endif
  
  bool left = (msg[6] == 'l');
  
#ifdef DEBUG
  if (!left && (msg[6] != 'r'))
    cout << "left/right key failed:" << msg << " <" << msg[6] << ">" << endl;
#endif

  int actor = atol(msg.substr(8).c_str());
  if (!left)
    actor += PlayerR_1 - 1;

#ifdef DEBUG
  cerr << "got player " << left << " " << actor << endl;
#endif
  return (SoccerActor)actor;
}

void usage(const char *name)
{
  cerr << "usage: " << name << " [-l logfile] [-s]" << endl;
}

int main(int argc, char **argv)
{
  Config::readFile("funsselrc");

  string logfile  = "soccerlog";

  char c;
  while ((c = getopt(argc, argv, "sl:")) != EOF) {
    switch (c) {
    case 'l':
      logfile  = optarg;
      break;
    case 's': {
      try {
	MsgStore store(logfile);
	store.show();
      }
      catch (Exception e) {
	cerr << "an exception occured: " << e << endl;
      }
      catch (...) {
	cerr << "an unexpected exception occured" << endl;
      }
      return 0;
    }
    default:
      usage(argv[0]);
    }
  }

  FilterSocket::enableInts();

#ifdef FORCE_CORE
  struct rlimit rlim;
  getrlimit(RLIMIT_CORE, &rlim);
  rlim.rlim_cur = rlim.rlim_max;
  cerr << "core dump enabled." << endl;  
  if (setrlimit(RLIMIT_CORE, &rlim) == -1)
    cerr << errno << endl;
#endif
  
  try {
    Base base("localhost", logfile);
#ifdef CONSOLE
    fcntl(1, F_SETFL, O_NONBLOCK);
    Socket sck(Config::getValueAsInt("proxy_socket"),
	       "localhost", MAX_SERVER_MSG_SIZE);
    sck.send(FILTER_STRING " (dispinit))");
    cout << ">";
    cout.flush();
#endif
    while (!base.termReq()) {
      if (!base.act()) {
#ifdef CONSOLE
	string str;
	while ((str = sck.poll()) != string())
	  cout << "recvd " << str << endl;
	int c = getchar();
	if (c != EOF) {
	  cout << endl;
	  switch (c) {
	  case '\n':
	    break;
	  case 's':
	    sck.send(FILTER_STRING " (server start))");
	    break;
	  case 'q':
	    sck.send(FILTER_STRING " (quit))");
	    break;
	  case 'p': {
	    int num;
	    if (scanf("%d\n", &num) != 1)
	      num = 1;
	    for (int i = 0; i < num; i++)
	      sck.send(FILTER_STRING " (poll))");
	    break;
	  }
	  case 'w': {
	    char buf[32];
	    scanf("%31s\n", buf);
	    sck.send(string(FILTER_STRING " (wind ") + buf + "))");
	    break;
	  }
	  case 'm': {
	    char buf[32];
	    scanf("%31s\n", buf);
	    sck.send(string(FILTER_STRING " (max_run ") + buf + "))");
	    break;
	  }
	  default:
	    usage(argv[0]);
	    break;
	  }
	  switch (base.getMode()) {
	  case Base::Running:
	    cout << "R ";
	    break;
	  case Base::Stopped:
	    cout << "S ";
	    break;
	  case Base::Playback:
	    cout << "P ";
	  }
	  cout << base.getTime() << " >";
	  cout.flush();
	}
#endif
	pause();
      }
    }
  }
  catch (Exception e) {
    cerr << "an exception occured: " << e << endl;
  }
  catch (...) {
    cerr << "an unexpected exception occured" << endl;
  }
  return 0;
}
