//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////
//
// mrbcoach.C - Online coach for MRB2000
//
// $Id: mrbcoach.C,v 1.41 2001/08/04 03:03:28 mrb Exp $
//
////////////////////////////////////////////////////////////////////////

#include "mrbcoach.h"
#include "snapshot.h"
#include "audiomsg.h"
#include "mrbtypes.h"
#include "minmax.h"

#include <strstream.h>
#include <string>

#include <stdlib.h>
#include <time.h>

//#define USEAUDIOTRACE
// #define USEANALYZETRACE
//#define USEACTTRACE
//#define USEDISTTRACE

//////////////////////////////////////////////////////////////////////
//
// some local constants to determine coach behaviour
//
//////////////////////////////////////////////////////////////////////

// number of messages to be left to allow second coach msgs.
// [C]
const int NUM_MSGS_SECOND_SAY= 30;
// max dist that is not assumed to be a goalie move (squared!)
const double MAX_GOALIE_DASH= sqr( PLAYER_SPEED_MAX*(1+QUANTIZE_STEP) );

// max error in player snapshot accepted without saying audio msgs
const double MAX_PLSNAP_DIST= 35;
// no of cycles in a standard situation before a second audio msg
const int AUMSG_BAN_TIME= 40;

#ifndef USEAUDIOTRACE
#define AUDIOTRACE(output)
#else
#define AUDIOTRACE(output) { ostrstream ost;\
 ost << "COACH AUDIO: " << output << (char)0;\
 Comm::sendMCmd(MLog(ost.str())); }
#warning OnlineCoach debug trace AUDIOTRACE is activated!
#endif

#ifndef USEANALYZETRACE
#define ANALYZETRACE(output)
#else
#define ANALYZETRACE(output) { ostrstream ost;\
 ost << "COACH analyze: " << output << (char)0;\
 Comm::sendMCmd(MLog(ost.str())); }
#warning OnlineCoach debug trace ANALYZETRACE is activated!
#endif

#ifndef USEACTTRACE
#define ACTTRACE(output)
#else
#define ACTTRACE(output) { ostrstream ost;\
 ost << "COACH act: " << output << (char)0;\
 Comm::sendMCmd(MLog(ost.str())); }
#warning OnlineCoach debug trace ACTTRACE is activated!
#endif

#ifdef USEDISTTRACE
#warning OnlineCoach debug trace DISTTRACE is activated!
#endif


MrbCoach::MrbCoach(const char *team, const char *host, const int port,
		   const bool logging, const char *configfile,
		   const bool FUNSSELmode) :
  Onlinecoach(team, host, port, logging, FUNSSELmode),
  sayMessageNow(false), lastMode(PLAY_MODE_UNKNOWN), 
  countOfMsgsLeft(SAY_COACH_CNT_MAX), thisGame(0),
  lastAumsg(-3 - AUMSG_BAN_TIME)
{
  // try to initialize config file
  if (!Config::readFile(configfile)) {
    cerr << "PANIC: Could not read config file: " << configfile << endl;
    exit(-1);
  }

  tactical     = new TacticalModule();

#ifdef RANDSUBS
  initSubs();
#endif
}

MrbCoach::~MrbCoach()
{
  delete tactical;

#ifdef RANDSUBS
  delete [] numPlType;
#endif
}

// acting in standard situations
void MrbCoach::act(Snapshot *snap) 
{
  ACTTRACE("Coach is acting.");
  
  // give additional messages to coach every 6000 cycles ...
  if (snap->time - thisGame >= HALF_TIME_IN_CYCLES * 2 ) {
    thisGame= snap->time;
    countOfMsgsLeft += SAY_COACH_CNT_MAX;
    // [C]
    ACTTRACE( "Added msg count for additional 6000 cycles. Have now left: " 
	  << countOfMsgsLeft );
  }
  
  if ( sayMessageNow ) {

    ACTTRACE( "Acting in a new standard situation ..." );
    sayMessageNow= false;

    lastAumsg= snap->time;
    
    if (countOfMsgsLeft == 0) {
      cerr << "Coach: Cannot scream: Lost voice earlier!" << endl;
    }
    
    AudioMsg coachMsg(0, snap->time, true);
    
    // this depends on player's worldfeedbackaudio() method !!
    coachMsg << true << *snap;
    coachMsg << tactical->getSnapshot()->otherPlayerType();

    if (!coachMsg.getState()) {
      cerr << "Coach : Fatal error: buffer overflow:" << coachMsg << endl;
    } else {
      say(coachMsg);
      
      AUDIOTRACE( "Sending msg: " << coachMsg );
      
      countOfMsgsLeft--;
      AUDIOTRACE( "Now coach has " << countOfMsgsLeft << " msgs left" );
    }
  }

#ifdef RANDSUBS
  doRandSubs(snap);
  printStats(snap);
#endif
}

//////////////////////////////////////////////////////////////////////
//
// analyze() - here the playmode is used to determine whether the 
//             coach should say something now or not.
//
//////////////////////////////////////////////////////////////////////

void MrbCoach::analyze(Snapshot *snap)
{ 
  tactical->feedbackWorld(snap);

  bool modeToAct(false);

  switch(snap->playMode) {
  case PLAY_MODE_KICK_OFF_US:
  case PLAY_MODE_KICK_OFF_THEM:
  case PLAY_MODE_GOAL_KICK_US:
  case PLAY_MODE_GOAL_KICK_THEM:
  case PLAY_MODE_DROP_BALL:
  case PLAY_MODE_OFFSIDE_US:
  case PLAY_MODE_OFFSIDE_THEM:
  case PLAY_MODE_KICK_IN_US:
  case PLAY_MODE_KICK_IN_THEM:
  case PLAY_MODE_CORNER_KICK_US:
  case PLAY_MODE_CORNER_KICK_THEM:
    modeToAct= true;
    break;
  case PLAY_MODE_FREE_KICK_US:
  case PLAY_MODE_FREE_KICK_THEM:
    if (lastMode != PLAY_MODE_OFFSIDE_THEM && 
	lastMode != PLAY_MODE_OFFSIDE_US ) {
      modeToAct= true;
    }
    break;
  default:
    // "normal" playmodes which cannot be handled by online coach
    
    // reset the counter for the last Audiomsg:
    lastAumsg= -2 - AUMSG_BAN_TIME;
  }

  
  // did the playmode change ?
  if (snap->playMode != lastMode) { 
    ANALYZETRACE( "Noted Change of playmode ...");
    lastMode= snap->playMode;
    if (modeToAct) {
      sayMessageNow=true;
      ANALYZETRACE( "Set flag: scream!" );
    } else {
      ANALYZETRACE( "Not right mode to act ...");
    }
  } else { // still same mode.
    if (modeToAct) {
      // get player's snapshots and compare them to the exact version
      // print out distances (pure, ball-related, player-related)

      AudioMsg aumsg;
      bool hasSnap;
    
      while (getNextMsg(aumsg)) {
	aumsg >> hasSnap;
	if (hasSnap) {
	  Snapshot plsnap;
	  aumsg >> plsnap;
	  
	  double plDist= Snapshot::dist(*snap, plsnap, aumsg.getSource());
      
	  if ((plDist > MAX_PLSNAP_DIST) &&
	      ((snap->time - lastAumsg) > AUMSG_BAN_TIME) )
	    sayMessageNow= true;
      
#ifdef USEDISTTRACE      
	  ostrstream ost;
	  ost << "COACH dists: for player "
	      << aumsg.getSource()
	      << " snap=" << Snapshot::dist(*snap, plsnap) 
	      << " ball=" << Snapshot::balldist(*snap, plsnap)
	      << " player=" << plDist;
	  if ((snap->time - lastAumsg) <= AUMSG_BAN_TIME)
	    ost << "BAN TIME";
	  ost << (char)0;
	  Comm::sendMCmd(MLog(ost.str()));
#endif
	}
      }
    }
  }
} 

// for testing of heterogenous player behaviour:
// random substitutions
#ifdef RANDSUBS

// some constants to change behaviour of coach during the game
// probability of a substitution during the game
#define SUBSPROB 0.3

// which step should be used for printing statistics?
#define STATALWAYS false
// if not always then only here:
#define STATSTEP 1

void MrbCoach::doRandSubs(const Snapshot *snap)
{
  play_mode pm(snap->playMode);
  if (pm == PLAY_MODE_BEFORE_KICK_OFF) {
    if (!initDone) {
      initDone= true;
      
      // don't try a sub for the goalie
      for (int i=1; i<11; i++) {
	// randomly choose a player type. Check if 
	// - standard player or
	// - hetero player and still some left of that type
	int pt=rand() % numPlTypes;      
	while ( pt != 0 && numPlType[pt-1] == 0 )
	  pt=rand() % numPlTypes; 
	doSub(i+1, pt);
      }
    }
  } else {
    if ((snap->time > 30) && (numSubs >0)) 
      if ((rand() % 100)/100.0 <= SUBSPROB) {
	numSubs--;
	// no sub for goalie, so quick hack here if goalie is assumed to 
	// be number 1 always
	int pl= rand() % 10;
	int pt=rand() % numPlTypes;      
	while ( pt != 0 && numPlType[pt-1] == 0 )
	  pt=rand() % numPlTypes; 
      	doSub(pl+2, pt);	
      } else {  { ostrstream ost;
      ost << "3 No change this time." << (char)0;
      Comm::sendMCmd(MLog(ost.str())); }
      }
  }
}

void MrbCoach::doSub(const int player, const int type)
{
  // change player i to type pt
  plType[player]= type;
  if (type != 0)
    numPlType[type-1]--;
  
  sendChangePlayer(player, type);
  
  { ostrstream ost;
  ost << "3 " << player << " set to type :" << type << (char)0;
  Comm::sendMCmd(MLog(ost.str())); }

  // inform about change of player type
  { ostrstream ost;
  ost << "HetCoach: " << player << " set to type :" << type << (char)0;
  Comm::sendMCmd(MMarker(Vector2d(player, type), ost.str())); }
}

void MrbCoach::initSubs()
{
  // number of subs left in Game (after KICK_OFF !)
  // set to 0 to avoid subs in online game
  numSubs= playerParams.subs_max;
  //  static unsigned int numSubs(0);
  
  // information about how many players of one type are left
  numPlTypes= playerParams.player_types;
  numPlType= new int[numPlTypes];
  for (int i=0; i<numPlTypes; i++)
    // [C] : Max number of players of one type on the field
    numPlType[i]= 3;

  srand(time(0));
  
  // initial subs not done yet!
  initDone= false;
}

void MrbCoach::printStats(const Snapshot *snap)
{
  if (!STATALWAYS && snap->time != STATSTEP)
    return;
  
  { ostrstream ost;
  ost << "0 Coach hetero player status:" << (char)0;
  Comm::sendMCmd(MLog(ost.str())); }
  
  for (int i=0; i<numPlTypes; i++) {
    ostrstream ost;
    ost << "1 " << (snap->playerType[i]) << (char)0;
    Comm::sendMCmd(MLog(ost.str()));       
  }
  
  for (int i=2; i<12; i++) {
    ostrstream ost;
    ost << "2 Player " << i << " has type " << plType[i] << (char)0;
    Comm::sendMCmd(MLog(ost.str()));       
  }
}

#endif
