//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// $Id: audiomsg.h,v 1.14 2001/07/03 08:32:50 goetz Exp $
//
// $manager: goetz$
//
// Audio Message with several string operations. 
//
//////////////////////////////////////////////////////////////////////

// [C]: Agent's audio capability is dependent on server implementation 

#ifndef _AUDIOMSG_H_
#define _AUDIOMSG_H_

#include <iostream.h>
#include <stdio.h>
#include <string.h>

#include "constants.h"
#include "coordinates.h"
#include "snapshot.h"

//////////////////////////////////////////////////////////////////////
//
// Constants
//
//////////////////////////////////////////////////////////////////////

const char DEFAULT_MSG_PREFIX[4]= "mrb";

//////////////////////////////////////////////////////////////////////
//
// Classes for handling simple data types and automatic encoding
// into "server-bytes"
//
//////////////////////////////////////////////////////////////////////

/** holds a 3-byte integer and its 3-byte encoding */

class Int3B 
{
public:
  /** produce 0 value */
  Int3B ();
  /** create with encoded string, should contain 3 byte encoding */
  Int3B (const char* encoded);
  /** create with integer value in 0..65536 otherwise 0 */
  Int3B (const unsigned int &val);
  void encode();
  void decode();
  unsigned int getValue() const { return value; };
  const char getEncoding(unsigned short pos) const 
  { return pos<3 ? encoding[pos] : 0; }
  operator int();
private:
  unsigned int value;
  char encoding[3];
}; // class Int3B

ostream & operator << (class ostream &os, const Int3B &i);


/** holds typical information about an object, encoded into 5 bytes:
 * coordinates x,y \in [-204.8, 204,7],
 * age \in [0..63] where age=63 is interpreted as eternal or invalid 
 */

class Val5B 
{
public:
  /** produce invalid entry */
  Val5B ();
  /** create with encoded string */
  Val5B (const char* encoded);
  /** create with x in -100.0..100.0, y in -100.0..100.0, age in 0..63
    invalid values lead to valid= false */
  Val5B (const double x_, const double y_, const unsigned int age_= 0,
	 const WorldObjectState state_= UNSAFE);
  
  Val5B (const Vector2d c, const unsigned int age_= 0,
	 const WorldObjectState state_= UNSAFE);
  
  /** get contents */
  double getX() const { return x; }; 
  double getY() const { return y; };
  unsigned int getAge() const { return age; };
  WorldObjectState getState() const { return state; };
  bool isValid() const { return (age < INVALIDAGE); };
  void encode();
  void decode();
  const char getEncoding(unsigned short pos) const
  { return pos<5 ? encoding[pos] : 0; }
  static const unsigned int INVALIDAGE;
  operator Vector2d();
private:
  double x;
  double y;
  unsigned int age;
  WorldObjectState state;
  char encoding[5];
}; // class Val5B

ostream &operator << (class ostream &os, const Val5B &v);

//////////////////////////////////////////////////////////////////////
//
// NodaChar: Holds one 6-bit char
//
//////////////////////////////////////////////////////////////////////

class NodaChar
{
public:
  NodaChar(const unsigned char val_=0);
  operator char() const;
private:
  unsigned char val;
}; // class NodaChar

ostream &operator << (class ostream &os, const NodaChar &c);

//////////////////////////////////////////////////////////////////////
//
// Class for collecting data to be sent / for receiving data
//
// If something is changed the checksums are NOT updated automatially!
//
//////////////////////////////////////////////////////////////////////

class AudioMsg 
{
public:
  /** create empty Msg */
  AudioMsg ();

  /** create message server time and a snapshot containing all information
   * and a free text message with length freetextlen
   */
  AudioMsg(const unsigned int destno,
	   const unsigned int stime,
	   const bool iscoachmsg= false);
  
  /** create an AudioMsg with a given encoding 
   */
  AudioMsg(const char *encoding);

  /** determine the type of the message (true=coachmsg=short) */
  bool isCoachMsg() const;

  /** set prefix of message */
  AudioMsg& setPrefix( const char prefix[4]);
  
  /** set side of player (l=true, r=false) */
  AudioMsg& setSourceAndSideIsLeft( const unsigned int playerno, 
				    const bool isLeft);
  /** set destination player no */
  AudioMsg& setDestination( const unsigned int playerno );
  /** set stime */
  AudioMsg& setServerTime( const unsigned int servertime);

  /** get src player no */
  unsigned int getSource() const;
  /** get dest player no */
  unsigned int getDestination() const;
  /** get server time */
  unsigned int getServerTime() const;
  /** get the side of the src player (left=true) */
  bool getSideIsLeft() const;

  
  /** verify the given prefix */
  bool verifyPrefix(const char prefix[4]) const;

  /** verify the dest-No; 0 is broadcast*/
  bool verifyDestination( const unsigned  int playerno ) const
  { return (getDestination() == 0) || (getDestination() == playerno); }

  /** create checksums for message */
  AudioMsg& createChkSums ();
  /** check if checksums are valid */
  bool isChkSumsValid () const;
  
  /** get the encoded version to transfer message 
   * don't even think about messing around with this pointer!!!!
   */
  const char *getEncoding() const { return encoded; };
  
  /**
   * functions for modifying the stream
   * set writing to pos 0 (write from beginning)
   * Note: this causes a rewindout as well and erases all data in stream !!
   */
  AudioMsg& rewindWrite();
  
  /** set read pos to 0 (read from beginning)
   */
  AudioMsg& rewindRead();

  /**
   * mark write stream
   */
  AudioMsg& markWrite();
  
  /**
   * rollback write stream to last marked position (if valid)
   * this resets the read position if neccessary!!
   */
  AudioMsg &rollbackWrite();

  /**
   * mark read stream
   */
  AudioMsg& markRead();

  /**
   * rollback read stream to last marked position (if valid)
   */
  AudioMsg &rollbackRead();

  /** get the state of the String (true if OK)
   */
  bool getState() const;

  /** clear status of the stream
   * Use only if REALLY necessary!
   */
  AudioMsg& clearState();

  /** get size of audiomsg (overall size) */
  unsigned int getSize() const;
  /** get no of chars left to read */
  unsigned int getReadLeft() const;
  /** get max no of chars which can be written */
  unsigned int getWriteLeft() const;
  

//////////////////////////////////////////////////////////////////////
//
// Operators for simple data types
//
//////////////////////////////////////////////////////////////////////
  
  AudioMsg& operator << (const Int3B &i);
  AudioMsg& operator >> (Int3B &i);
  
  AudioMsg& operator << (const Val5B &v);
  AudioMsg& operator >> (Val5B &v);
  
  AudioMsg& operator << (const bool &b);
  AudioMsg& operator >> (bool &b);
  
  AudioMsg& operator << (const char &c);
  AudioMsg& operator >> (char &c);
  
  /** precision of angle: 2pi/4096 (about 1.5e-3) */
  AudioMsg& operator << (const Angle &a);
  AudioMsg& operator >> (Angle &a);
  
  /** the double operator << takes values in [-200.0, 200.0] only !! */
  AudioMsg& operator << (const double &d);
  AudioMsg& operator >> (double &d);

  AudioMsg& operator << (const NodaChar &c);
  AudioMsg& operator >> (NodaChar &c);
  
private:
  unsigned int getWritePos() const;
  /** set position of stream input; returns position or 0 if invalid */
  bool setWritePos(unsigned int pos);
  unsigned int getReadPos() const;
  /** set position of stream output; returns position or 0 if invalid */
  bool setReadPos(unsigned int pos);
  unsigned int chksum() const;
  char encoded[MAX_SAY_MSG_SIZE];
  bool state;

  unsigned int MSG_LENGTH;

  unsigned int readMark;
  unsigned int writeMark;
}; // class AudioMsg

class ostream & operator << (class ostream &os, const AudioMsg &a);

#endif
