//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2001 competition
//
// Copyright (C) 1998, 1999, 2000, 2001 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program is published under the terms of the GNU General Public
// License as published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: sensor.h,v 1.15 2001/03/22 10:42:10 satan Exp $
// 
//  sensor.h - Declaration of the SensorData - type
// 
//////////////////////////////////////////////////////////////////////


#ifndef __SENSOR_H_
#define __SENSOR_H_
#include "Ensure.h"
#include "constants.h"
#include "mrbtypes.h"




// types of clients that can check into the server
enum client_type {
  CLIENT_UNKNOWN = -1,
  CLIENT_GOALIE,
  CLIENT_FIELDPLAYER,
  CLIENT_ONLINECOACH,
  CLIENT_TRAINCOACH
};

enum sensor_data_type {
  SENSOR_TYPE_SEE,
  SENSOR_TYPE_HEAR,
  SENSOR_TYPE_SENSE_BODY,
  SENSOR_TYPE_EXACT,
  SENSOR_TYPE_OK,
  SENSOR_TYPE_ERROR,
  SENSOR_TYPE_PLAYER_TYPE,
  SENSOR_TYPE_PLAYER_CHANGED
};

enum team_number {
  TEAM_UNKNOWN = -1,
  TEAM_US,
  TEAM_THEM
};

enum goalie_player {
  GOALIE_UNKNOWN = -1,
  GOALIE_TRUE,
  GOALIE_FALSE
};

enum sender_type {
  SENDER_SELF = 0,
  SENDER_REFEREE,
  SENDER_DIR,
  SENDER_PLAYER,
  SENDER_ONLINECOACH,
  SENDER_COACH
};

// structure for synchronous player commands for one simulation cycle
// see Soccerserver Manual Appendix A2
// (asynchronous commands are sent directly using their own send-methods)

enum CycleCommand {
  NO_COMMAND =  0,    // param1 = param2 = 0
  KICK_COMMAND,       // param1 = power, param2 = dir_rad
  DASH_COMMAND,       // param1 = power, param2 = 0
  TURN_COMMAND,       // param1 = moment_rad, param2 = 0
  CATCH_COMMAND,      // for goalie only, param1 = dir_rad, par2 = 0
  MOVE_COMMAND,       // pm_before_kick_off only, param1=x, param2=y
  CHANGE_VIEW_COMMAND // param1 = view_width, param2 = view_quality
}; 

// the structure that stores the commands together with their parameters
struct CycleAction {

  int time; // the timestep during which the command was sent to the server
  
  CycleCommand command; // the command sent
  
  double param1;    // params of the command, see above at command description
  double param2;
};

// the struct that stores each playertype
struct PlayerType {
  int player_type;
  double player_speed_max;
  double stamina_inc_max;
  double player_decay;
  double inertia_moment;
  double dash_power_rate;
  double player_size;
  double kickable_margin;
  double kick_rand;
  double extra_stamina;
  double effort_max;
  double effort_min;
};

// Distances/directions in a SensorData object may be invalid, i.e.
// we may only see the distance but not the direction etc.
// These constants are used to mark distances or directions as invalid
const double INVALID_DIST = -10000;
const double INVALID_DIR = -10000;

inline bool IsDistValid(double dist) {return dist > INVALID_DIST;}
inline bool IsDirValid(double dir) {return dir > INVALID_DIR;}




/*
 *  Components of the SensorData class
 */

struct SensorObject {
  bool valid;		// true:  object seen
  // false: object not seen, entries are invalid
  double dist;
  double dir;		// In radians
  double dist_change;
  double dir_change;	// In radians
};

struct SensorPlayer : public SensorObject {
  team_number   team;
  int           number;		// -1 = unknown
  goalie_player goalie;
  double        face_dir;      // in radians
  double        body_dir;      // in radians
};

struct SensorGoal : public SensorObject {
  goal_type which;	// Which goal is it? (left/right)
  team_number team;	// Is it our goal or the enemy's?
};

struct SensorFlag : public SensorObject {
  flag_type which;
};

struct SensorLine : public SensorObject {
  line_type which;
};

struct ExactObject {
  bool valid;		// true:  object seen
  // false: object not seen, entries are invalid
  double x;  // position
  double y;
  double x_vel; // velocities in direction of axes
  double y_vel;
};

struct ExactPlayer : public ExactObject {
  team_number   team;
  int           number;
  goalie_player goalie;
  double        face_dir;      // in radians
  double        body_dir;      // in radians
};

struct ExactGoal : public ExactObject {
  goal_type which;	// Which goal is it? (left/right)
  team_number team;	// Is it our goal or the enemy's?
};


/*
 *  The SensorData class holds sensoric information obtained from the server
 */


struct SensorData {

  sensor_data_type type; // type of sensor data received

  int time;		// timestamp of data from server


  // if type == SENSOR_TYPE_OK or type == SENSOR_TYPE_ERROR, 
  // none of the following entries is valid, and the time is set to -1.
  // These SensorDatas are not yet used by any of the components of an agent.

  // the following valid if type == SENSOR_TYPE_PLAYER_TYPE
  PlayerType playerType;

  // the following valid if type == SENSOR_TYPE_SEE
  SensorObject ball;
  SensorGoal goal[2];
  SensorPlayer player[21];
  SensorFlag flag[NUM_FLAGS];
  SensorLine line[NUM_LINES];

  // for type = SENSOR_TYPE_PLAYER_CHANGED
  int number;
  bool ourTeamChanged;   // true if a player of our team was changed
  int heteroType;        // currently UNKNOWN if player is in other team

  // for type == SENSOR_TYPE_HEAR
  sender_type sender;                 // always valid when S_T_HEAR
  CoachMsgType coachMsgType;          // always valid when S_T_HEAR
  char message[MAX_SAY_MSG_SIZE];     // always valid when S_T_HEAR
  play_mode mode;		      // only valid if sender == SENDER_REFEREE
  int score;                          // only valid if mode==GOAL_US||GOAL_THEM
  double sender_dir;	              // only valid if sender == SENDER_DIR
  char sender_name[MAX_SAY_MSG_SIZE]; // only valid if sender == SENDER_PLAYER

  // the following valid if type == SENSOR_TYPE_SENSE_BODY
  view_width my_view_width;
  view_quality my_view_quality;
  double stamina;
  double effort;
  double speed;        // radius of speed vector
  double speed_angle;  // angle of speed vector, as given by server
  double head_angle;
  int kick_count;
  int dash_count;
  int turn_count;
  int say_count;
  int turn_neck_count;
  int catch_count;
  int move_count;
  int change_view_count;

  // the following valid if type == SENSOR_TYPE_EXACT
  ExactObject exactball;
  ExactPlayer exactplayer[22];
  ExactGoal   exactgoal[2];

};



#endif
